# The command used to play streams
# You can add arguments to the command a la Python's subprocess module
PLAYER = ['/usr/bin/mpv']

# Respect default terminal colors in curses interface.
DEFAULT_TERMINAL_COLORS = False



import curses

import textwrap
import logging
import re
import os

from collections import namedtuple

BULLET = '•'

MIN_HEIGHT = 5

CTRL = 0b1100000
DOWN_KEYS = (curses.KEY_DOWN, ord('j'))
UP_KEYS = (curses.KEY_UP, ord('k'))
START_KEYS = (curses.KEY_HOME, ord('g'))
PAGE_UP_KEYS = (curses.KEY_PPAGE, CTRL ^ ord('u'))
PAGE_DOWN_KEYS = (curses.KEY_NPAGE, CTRL ^ ord('d'))
END_KEYS = (curses.KEY_END, ord('G'))
QUIT_KEYS = (27, ord('q'))
PAGE_STEP = 15

starting_space = re.compile(r'^ +')

# Set shorter delay for Esc key
os.environ.setdefault('ESCDELAY', '25')

class Colors:
    pass

COLORS = Colors()

def wrap(text, width, indent=0, **kwargs):
    try:
        ind = ' ' * indent
        first, *rest = text.splitlines()
        first_wrap = textwrap.wrap(first, width, subsequent_indent=ind, **kwargs)
        rest_wrap = []
        for line in rest:
            match = starting_space.match(line)
            extra_indent = ''
            if match:
                extra_indent = match.group()
            rest_wrap.extend(textwrap.wrap(line, width, subsequent_indent=ind + extra_indent, initial_indent=ind, **kwargs))
        return [*first_wrap, *rest_wrap]
    except ValueError:
        return []

class Line:
    def __init__(self, win, text, width, itemize):
        self.text = text
        self.width = width - 4
        indent = 0
        if itemize:
            text = f'{itemize} {text}'
            indent = len(itemize) + 1
        self.lines = wrap(text, self.width, indent=indent, max_lines=3)
        self.size = len(self.lines)
        self.win = win

    def display(self, index, selected):
        color = COLORS.normal
        if selected:
            color = COLORS.highlight

        for line in self.lines:
            self.win.addstr(index, 2, line.ljust(self.width), color)
            index += 1
        return index


class Page:
    def __init__(self, lines):
        self.lines = lines
        self.index = 0

    def up(self):
        if self.index == 0:
            return False
        self.index -= 1
        return True

    def down(self):
        if self.index == len(self.lines) - 1:
            return False
        self.index += 1
        return True

    def last(self):
        self.index = len(self.lines) - 1

    def first(self):
        self.index = 0

    def display(self):
        display_index = 1
        for ind, line in enumerate(self.lines):
            display_index = line.display(display_index, ind == self.index)

Choice = namedtuple('Choice', ['index', 'item'])

class ListDisplay:
    def __init__(self, items, title=None, display=str, itemize=None, initial_index=0, allow_exit=True, default_terminal_colors=False):
        if not items:
            raise ValueError('List cannot be empty')
        if initial_index >= len(items):
            raise ValueError('Initial index too large')

        self.index = initial_index
        self.page = 0
        self.title = title
        self._setup(default_terminal_colors)
        self.items = items
        self.display = display
        self.itemize = itemize
        self.allow_exit = allow_exit

        self._update_lines()
        self._paginate()
        self._find_current_page()

    @property
    def current_page(self):
        return self.pages[self.page]

    @property
    def title_height(self):
        if self.title is None or not self.title_lines:
            return 0
        return len(self.title_lines)

    def _paginate(self):
        self.pages = []
        page_lines = []
        page_height = 0
        for index, line in enumerate(self.lines):
            if page_height + line.size >= self.rows - 1:
                self.pages.append(Page(page_lines))
                page_lines = []
                page_height = 0
            page_lines.append(line)
            page_height += line.size
        self.pages.append(Page(page_lines))

    def _update_size(self):
        y, x = self.screen.getmaxyx()
        self.rows = y - 2
        self.cols = x - 2

    def _setup(self, default_colors):
        self.screen = curses.initscr()
        self.screen.erase()
        curses.noecho()
        curses.cbreak()
        curses.start_color()

        if default_colors:
            curses.use_default_colors()
            COLORS.normal = curses.A_NORMAL
            COLORS.highlight = curses.A_STANDOUT
            COLORS.title = curses.A_BOLD
        else:
            curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
            curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_WHITE)
            curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
            COLORS.normal = curses.color_pair(1)
            COLORS.highlight = curses.color_pair(2)
            COLORS.title = curses.color_pair(3) | curses.A_BOLD

        self.screen.keypad(1)
        curses.curs_set(0)
        self._update_size()

        self._setup_title()
        self.box = curses.newwin(self.rows, self.cols, self.title_height + 1, 1)
        self.box.attron(COLORS.normal)
        self.box.box()
        self.box.attroff(COLORS.normal)

    def _update_lines(self):
        self.lines = [Line(self.box, self.display(it), self.cols, self.itemize) for it in self.items]

    def _setup_title(self):
        if self.title is None:
            self.title_lines = []
            return
        self.title_lines = wrap(self.title, self.cols - 2, max_lines=5)
        self.rows -= len(self.title_lines)

    def _display_title(self):
        for index, line in enumerate(self.title_lines):
            self.screen.addstr(index + 1, 2, line, COLORS.title)

    def _find_current_page(self):
        ind = self.index
        for page_ind, page in enumerate(self.pages):
            page_lines = len(page.lines)
            if ind >= page_lines:
                ind -= page_lines
            else:
                self.page = page_ind
                page.index = ind
                return

    def _resize(self):
        self._update_size()
        if self.rows < MIN_HEIGHT:
            return
        self._setup_title()
        self.box = curses.newwin(self.rows, self.cols, self.title_height + 1, 1)
        self._update_lines()
        self.screen.erase()
        self.box.erase()
        self._paginate()
        self._find_current_page()

    def _display_page_number(self):
        index = self.rows - 1
        page_text = f' Page {self.page + 1} / {len(self.pages)} '[:self.cols-2]
        position = max(self.cols - len(page_text) - 2, 1)
        self.box.addstr(index, position, page_text, COLORS.normal)

    def _display_page(self):
        if self.rows < MIN_HEIGHT:
            return
        self.current_page.display()
        self._display_page_number()
        self._display_title()

        self.screen.refresh()
        self.box.refresh()

    def page_up(self):
        if self.index < PAGE_STEP:
            self.index = 0
        else:
            self.index -= PAGE_STEP
        self._find_current_page()

    def page_down(self):
        if self.index > len(self.lines) - 1 - PAGE_STEP:
            self.index = len(self.lines) - 1
        else:
            self.index += PAGE_STEP
        self._find_current_page()

    def up(self):
        if self.index == 0:
            return
        self.index -= 1
        if not self.current_page.up():
            self.page -= 1
            self.current_page.last()

    def down(self):
        if self.index == len(self.lines) - 1:
            return
        self.index += 1
        if not self.current_page.down():
            self.page += 1
            self.current_page.first()

    def last(self):
        self.index = len(self.lines) - 1
        self._find_current_page()

    def first(self):
        self.index = 0
        self._find_current_page()

    def handle_keypress(self, x):
        actions = [
            (UP_KEYS, self.up),
            (DOWN_KEYS, self.down),
            (PAGE_UP_KEYS, self.page_up),
            (PAGE_DOWN_KEYS, self.page_down),
            ((curses.KEY_RESIZE,), self._resize),
            (START_KEYS, self.first),
            (END_KEYS, self.last),
        ]
        if x == ord('\n'):
            return False
        for keys, action in actions:
            if x in keys:
                action()
        return True

    def choose(self):
        try:
            self._display_page()

            x = self.screen.getch()
            while x not in QUIT_KEYS or not self.allow_exit:
                if not self.handle_keypress(x):
                    break

                self.box.erase()
                self.box.attron(COLORS.normal)
                self.box.border(0)
                self.box.attroff(COLORS.normal)

                self._display_page()

                x = self.screen.getch()
            else:
                self.index = None

            curses.endwin()

            if self.index is None:
                return None
            return Choice(self.index, self.items[self.index])
        except:
            logging.exception('Something bad happened')


def choose(choices, title=None, display=None, index=True, **args):
    if 'bullet' in args:
        args['itemize'] = BULLET
        del args['bullet']
    list_display = ListDisplay(
            choices,
            title,
            display=display,
            **args
    )
    return list_display.choose()



import sys
from pathlib import Path

CONFIG_DIR = Path.home() / '.config' / 'ruvcli'
CONFIG_PATH = CONFIG_DIR / 'config.py'

sys.path.append(str(CONFIG_DIR))

from .default_config import *
if CONFIG_PATH.exists():
    from config import *

def config_exists():
    return CONFIG_PATH.exists()

def copy_config():
    if not CONFIG_DIR.exists():
        CONFIG_DIR.mkdir(parents=True)
    default_config = (Path(__file__).parent / 'default_config.py').read_text()
    CONFIG_PATH.write_text(default_config)




import argparse
import subprocess
import re
import sys
from functools import partial
from datetime import timedelta, date
from urllib.parse import urlsplit, parse_qs

from requests.exceptions import HTTPError, ConnectionError

from .choose import choose
from .conf import config_exists, copy_config, CONFIG_PATH, PLAYER, DEFAULT_TERMINAL_COLORS
import ruv.api as api
import ruv.__version__ as about
from .geoapi import get_channel_stream

eprint = partial(print, file=sys.stderr)
choose = partial(choose, default_terminal_colors=DEFAULT_TERMINAL_COLORS)

RUV_URL = 'http://ruv.is/'
CHANNEL_NAMES = ['ruv', 'ruv2']
RADIO_NAMES = ['ras1', 'ras2', 'rondo']
RADIO_ALIASES = {'rondo': 'ras3'}


def graceful(func):
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except HTTPError as e:
            eprint(f'The RUV API responded with an error ({e.response.status_code})')
            eprint(f'URL: {e.request.url}')
        except ConnectionError:
            eprint('Error connecting to the RUV API')
    return wrapper


@graceful
def play_stream(args, channel):
    res = get_channel_stream(channel)
    if res.get('geoblock'):
        eprint('You appeared to be geoblocked')
    url = res.get('url')
    if not url:
        eprint(f"RUV did not provide a URL for live streaming channel '{channel}'")
        return

    if args and args.video_player:
        player = [args.video_player]
    else:
        player = PLAYER
    subprocess.call(player + [url])


def live(args):
    play_stream(args, args.channel)


def default_live():
    play_stream(None, 'ruv')


def default_live2():
    play_stream(None, 'ruv2')


def radio(args):
    play_stream(args, RADIO_ALIASES.get(args.channel, args.channel))


def menu(choices, title, on_chosen, display=lambda x: x.display()):
    choice = None
    while True:
        index = (choice and choice.index) or 0
        choice = choose(
                choices,
                title=title,
                display=display,
                initial_index=index
        )
        if choice is None:
            break
        on_chosen(choice.item)


def program_details_menu(args, program_id):
    details = api.program_details(program_id)
    menu(details.episodes, details.header, lambda ep: play_stream(args, ep.file))


def choose_program_menu(args, programs, title):
    def when_chosen(prog):
        if not prog.multiple_episodes:
            play_stream(args, prog.episodes[0].file)
        else:
            program_details_menu(args, prog.id)
    menu(programs, title, when_chosen)


@graceful
def play(args):
    parsed = urlsplit(args.url)
    parts = parsed.path.split('/')
    prog_id = (parts or '') and parts[-1]
    query = parse_qs(parsed.query)
    if not re.fullmatch('\d+', prog_id) or 'ep' not in query:
        print('Unplayable URL')
        return
    ep_id = (query['ep'] or '') and query['ep'][0]
    details = api.program_details(prog_id)
    eps = [ep for ep in details.episodes if ep.id == ep_id]
    if not eps:
        print('Episode not found')
        return
    play_stream(args, eps[0].file)



@graceful
def search(args):
    results = api.search(args.query)
    if results.empty():
        print('No shows matched query')
        return
    if args.play:
        program = results.programs[0]
        print('Chosen show is: %s' % program.title)
        episodes = program.episodes
        if episodes:
            if args.offset:
                details = api.program_details_menu(program.id)
                episodes = details.episodes
            if args.offset >= len(episodes):
                print('Offset out of range, playing oldest')
                args.offset = -1
            episode = episodes[args.offset]
            print("Playing:")
            if program.multiple_episodes:
                print(episode.display())
            else:
                print(program.display())
            play_stream(args, episode.file)
        else:
            print('No episodes available')
    else:
        choose_program_menu(args, results.programs, results.title)


@graceful
def featured(args):
    feat = api.featured()
    menu(feat.panels, 'Featured programs', lambda pan: choose_program_menu(args, pan.programs, pan.title), display=lambda pan: pan.title)


@graceful
def schedule(args):
    def when_selected(event):
        if event.is_playable():
            play_stream(args, event.program.episodes[0].file)

    day = date.today() + timedelta(days=args.day)
    schedule = api.schedule(args.channel, day)
    if not schedule.events:
        print('No schedule for selected day')
        return
    menu(schedule.events, schedule.long_title, when_selected)


def config(args):
    if config_exists():
        inp = input('Config file already exists. Overwrite? [y/N] ')
        if not inp.lower().startswith('y'):
            return
    copy_config()
    print(f"Config copied to '{CONFIG_PATH}'")


def version():
    print(f'{about.__name__} {about.__version__}')
    print(f'License: {about.__license__}')
    print(f'Written by {about.__author__} ({about.__author_email__})')


def main():
    parser = argparse.ArgumentParser(description='A command line interface for RUV', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-p', '--video-player', metavar='PLAYER', help='The video player used to play the stream', default=None)
    parser.add_argument('--version', help='Print the version information and exit', action='store_true')

    subparsers = parser.add_subparsers()

    live_parser = subparsers.add_parser('live', help='Watch RUV live')
    live_parser.add_argument('channel', metavar='CHANNEL', help=f'Channel to stream. Choose from: {", ".join(CHANNEL_NAMES)}', default='ruv', nargs='?', choices=CHANNEL_NAMES)
    live_parser.set_defaults(func=live)

    radio_parser = subparsers.add_parser('radio', help='Listen to live radio')
    radio_parser.add_argument('channel', metavar='CHANNEL', help=f'Radio channel to stream. Choose from: {", ".join(RADIO_NAMES)}', default='ras2', nargs='?', choices=RADIO_NAMES)
    radio_parser.set_defaults(func=radio)

    schedule_parser = subparsers.add_parser('schedule', help='See channel schedules')
    schedule_parser.add_argument('-c', '--channel', metavar='CHANNEL', help=f'Channel to stream. Choose from: {", ".join(CHANNEL_NAMES)}. Default: %(default)s', default='ruv')
    schedule_parser.add_argument('day', metavar='DAY', nargs='?', default=0, type=int, help='Day offset. 0 is today, -n is n days in the past and n is n days in the future.')
    schedule_parser.set_defaults(func=schedule)

    show_parser = subparsers.add_parser('search', help='Search for programs')
    show_parser.add_argument('query', metavar='QUERY', help='Search shows matching this query')
    show_parser.add_argument('-p', '--play', help='If query matches any program, the latest episode of the first program will be played.', action='store_true')
    show_parser.add_argument('-o', '--offset', help='Offset when playing an episode. An offset of 1 means the second latest episode will be played, 2 the third latest, etc.',
            default=0, type=int, metavar='OFFSET')
    show_parser.set_defaults(func=search)

    featured_parser = subparsers.add_parser('featured', help='List features programs')
    featured_parser.set_defaults(func=featured)

    play_parser = subparsers.add_parser('play', help='Play a program from a URL')
    play_parser.add_argument('url', metavar='URL', help='URL of the page containing the stream')
    play_parser.set_defaults(func=play)

    config_parser = subparsers.add_parser('config', help='Copy the default configuration to your home directory for customization')
    config_parser.set_defaults(func=config)

    args = parser.parse_args()
    if args.version:
        version()
        return
    if not hasattr(args, 'func'):
        parser.print_help()
    else:
        args.func(args)

if __name__ == '__main__':
    main()



import os

from setuptools import setup

ROOT = os.path.abspath(os.path.dirname(__file__))

with open('README.md') as f:
    readme = f.read()

about = {}
with open(os.path.join(ROOT, 'ruv', '__version__.py')) as f:
    exec(f.read(), about)

setup(
    name=about['__name__'],
    version=about['__version__'],
    description='Command line interface for RUV (http://www.ruv.is)',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/hjalti/ruv-cli',
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__version__'],
    packages=['ruv'],
    python_requires='>=3.6.0',
    install_requires=[
        "requests",
    ],
    entry_points={
        'console_scripts': [
            'ruv=ruv:main',
            'ruv-live=ruv:default_live',
            'ruv2-live=ruv:default_live2',
        ],
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ]
)



import requests
from functools import wraps
from urllib.parse import urljoin
from datetime import date
from .models import Overview, SearchResults, ProgramDetails, Schedule

API_URL = 'https://api.ruv.is/api/'

def json(model):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            resp = func(*args, **kwargs)
            resp.raise_for_status()
            return model(resp.json())
        return wrapper
    return decorator

def api_path(path):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(urljoin(API_URL, path), *args, **kwargs)
        return wrapper
    return decorator

@json(SearchResults)
@api_path('programs/search/tv/')
def search(path, search_str):
    search_path = path + search_str
    return requests.get(search_path)

@json(Overview)
@api_path('programs/featured/tv/')
def featured(path):
    return requests.get(path)

@json(ProgramDetails)
@api_path('programs/program/%s/all/')
def program_details(path, program_id):
    path = path % program_id
    return requests.get(path)

@json(Schedule)
@api_path('schedule/%s/%s/')
def schedule(path, channel='ruv', day=None):
    if day is None:
        day = date.today()
    return requests.get(path % (channel, date.strftime(day, '%Y-%m-%d')))

@json(SearchResults)
@api_path('programs/category/tv/')
def category(path, category):
    path = path + category
    return requests.get(path)



from pprint import pformat
from datetime import datetime
import textwrap

TIME_FORMAT = '%Y-%m-%dT%H:%M:%S'

class ModelBase:
    def __init__(self, dic):
        self.__dict__.update(dic)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return pformat(self.__dict__)

    def _expand_list(self, name, model):
        if hasattr(self, name):
            attr = getattr(self, name)
            if attr:
                setattr(self, name, [model(a) for a in attr])

    def _expand(self, name, model):
        if hasattr(self, name):
            attr = getattr(self, name)
            if attr:
                setattr(self, name, model(attr))


def format_description(lines):
    desc = []
    for line in lines:
        desc.append(textwrap.indent(line, 4 * ' '))
    if not desc:
        desc.append('    [No description]')
    return '\n'.join(desc)


def parse_date(date):
    return datetime.strptime(date, TIME_FORMAT)


def format_time(date):
    return datetime.strftime(date, '%H:%M')


class Episode(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self.files = ModelBase(self.files)

    def display(self):
        return '\n'.join([
            f'{self.title} ({self.firstrun})',
            format_description((self.short_description or '').strip().splitlines())
        ])


class Program(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self._expand_list('episodes', Episode)
        if not self.title:
            self.title = 'unknown'

    def _title(self):
        if not self.foreign_title:
            return self.title
        return f'{self.title} ({self.foreign_title})'

    def display(self):
        return '\n'.join([
            f"{self._title().strip()} ({'Series' if self.multiple_episodes else 'Movie/Short'})",
            format_description((self.short_description or '').strip().splitlines())
        ])


class Panel(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self._expand_list('programs', Program)

    def display(self, indent=0):
        programs = '\n'.join(p.display(indent=1) for p in self.programs)
        return '\n'.join([
            f'{self.title.upper()}:',
            programs
        ])


class SearchResults(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self._expand_list('programs', Program)

    def empty(self):
        return self.program_count == 0

    def __bool__(self):
        return not self.empty()

    @property
    def title(self):
        return f'Search results for "{self.search_query}"'


    def display(self):
        programs = '\n'.join(p.display(indent=1) for p in self.programs)
        return '\n'.join([
            f'{self.title}:',
            programs
        ])


class ProgramDetails(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self._expand_list('episodes', Episode)
        self._expand_list('panels', Panel)

    @property
    def long_description(self):
        if not self.description:
            return '[No description]'
        return '\n'.join(self.description)

    def _title(self):
        if not self.title:
            return '[no title]'
        if not self.foreign_title:
            return self.title
        return f'{self.title} ({self.foreign_title})'

    @property
    def header(self):
        return '\n'.join([
            f"{self._title().upper()}",
            self.long_description
        ])


class Overview(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self._expand_list('panels', Panel)

    def display(self, indent=0):
        return '\n'.join(p.display() for p in self.panels)


class Event(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self._expand('program', Program)

    @property
    def start_time_friendly(self):
        return format_time(parse_date(self.start_time))

    @property
    def full_description(self):
        return '\n'.join(self.description)

    def is_available(self):
        return self.web_accessible and self.program is not None

    def has_passed(self):
        return self.program and self.program.episodes

    def is_playable(self):
        return self.has_passed() and self.is_available()

    def _episode_of(self):
        if hasattr(self, 'episode_number'):
            return f' ({self.episode_number} of {self.number_of_episodes})'
        return ''

    def _title(self):
        if self.original_title:
            return f'{self.title} ({self.original_title})'
        return self.title

    def display(self, indent=0):
        availability = ''
        if not self.is_available():
            availability = ' [Not available]'
        elif not self.has_passed():
            availability = ' [Not started]'

        return '\n'.join([
            f'{self.start_time_friendly}: {self._title()}{self._episode_of()}{availability}',
            format_description(self.description)
        ])


class Schedule(ModelBase):
    def __init__(self, dic):
        super().__init__(dic)
        self._expand_list('events', Event)

    @property
    def long_title(self):
        return f'{self.title} - {self.selected_date}'

    def display(self):
        events = '\n'.join(p.display(indent=1) for p in self.events)
        return '\n'.join([
            self.long_title,
            events
        ])
